#!/usr/bin/env python3

"""
kr_json: functions for translating keyrings and keys to xml and vice versa
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

import locale
from datetime import datetime
#import xml.etree.ElementTree as ET
import json

from .kr_storage import \
    connection_open, create_keyring, \
    get_all_keyrings, search_keys, NotAvailableException, \
    LockedException, KeyNotFoundException, PromptDismissedException

# FIXME: needs to move to util.py ...!!
def timestamp(since_epoch):
    """returns a string according to local locale
    based on seconds since epoch 1-JAN-1970"""
    if since_epoch:
        from_timestamp = datetime.fromtimestamp(since_epoch)
        date_string = from_timestamp.strftime(
            locale.nl_langinfo(locale.D_T_FMT))
        return date_string
    return""


#def dic_key( dic, key):
def dic_key(key):
    """ materialized key """
    k_label = key.get_label()
    """
    dic[k_label] = {}
    dic[k_label]["created"] = str(int(key.get_created()))
    dic[k_label]["modified"] = str(int(key.get_modified()))
    dic[k_label]["secret"] = key.get_secret().decode("utf-8")
    dic[k_label]["attributes"] = key.get_attributes()
    """
    dic = {}
    dic["key_label"] = k_label
    dic["created"] = str(int(key.get_created()))
    dic["modified"] = str(int(key.get_modified()))
    dic["secret"] = key.get_secret().decode("utf-8")
    dic["attributes"] = key.get_attributes()
    return dic


#def dic_keyring( dic, keyring):
def dic_keyring( keyring):
    """ materialized keyring """
    kr_label = keyring.get_label()
    dic = {}
    dic["keyring_label"] = kr_label
    dic["keys" ] = []
    for key in keyring.get_all_keys():
        dic["keys" ].append(dic_key(key))
    return dic


def dic_all_keyring():
    """ returns the dictionary representation of all keyrings """

    dic = {}
    dic["keyrings"] = []
    conn = connection_open()
    for keyring in get_all_keyrings(conn):
        if keyring.is_locked():
            keyring.unlock()
        if not keyring.is_locked() and len(keyring.get_label()) > 0:
            dic["keyrings"].append(dic_keyring(keyring))
    return dic

def dic2str(kr_dic):
    return json.dumps(kr_dic)

def str2dic(kr_str):
    return json.loads( kr_str )




"""
JSON format:

json_str = '{ "keyring" : [ {  "keyring label" : "<keyring label>", "key" : [ { "key label" : "<key label>" }, {  "key label" : "<key2 label>" } ] } , { "keyring label" : "<keyring label>" } ] }'
"""



#!/usr/bin/env python3

"""
kr_time: functions for time manipulation
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

import locale
from datetime import datetime


# remap some datetime functions
now = datetime.now
#timedelta = datetime.timedelta



def timestamp(since_epoch):
    """returns a string according to local locale
    based on seconds since epoch 1-JAN-1970
    if since_epoch is none, an empty string is returned"""
    if since_epoch:
        from_timestamp = datetime.fromtimestamp(since_epoch)
        date_string = from_timestamp.strftime(
            locale.nl_langinfo(locale.D_T_FMT))
        return date_string
    return ""







#!/usr/bin/env python3

"""
kr_memstorage: an in--memory storage provider, used for testing 
purposes and intermediairy layer for backup/restore import/export etc
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

from datetime import datetime
from typing import Dict, Iterator
import locale
from datetime import datetime
#import xml.etree.ElementTree as ET
import json

DEFAULT_COLLECTION = "????"

#from kr_pass import \
from .kr_secretstorage import \
    connection_open, create_keyring, \
    get_all_keyrings, search_keys, NotAvailableException, \
    LockedException, KeyNotFoundException, PromptDismissedException

# FIXME: needs to move to util.py ...!!
def timestamp(since_epoch):
    """returns a string according to local locale
    based on seconds since epoch 1-JAN-1970"""
    if since_epoch:
        from_timestamp = datetime.fromtimestamp(since_epoch)
        date_string = from_timestamp.strftime(
            locale.nl_langinfo(locale.D_T_FMT))
        return date_string
    return""


#def dic_key( dic, key):
def dic_key(key):
    """ materialized key """
    k_label = key.get_label()
    """
    dic[k_label] = {}
    dic[k_label]["created"] = str(int(key.get_created()))
    dic[k_label]["modified"] = str(int(key.get_modified()))
    dic[k_label]["secret"] = key.get_secret().decode("utf-8")
    dic[k_label]["attributes"] = key.get_attributes()
    """
    dic = {}
    dic["key_label"] = k_label
    dic["created"] = str(int(key.get_created()))
    dic["modified"] = str(int(key.get_modified()))
    dic["secret"] = key.get_secret().decode("utf-8")
    dic["attributes"] = key.get_attributes()
    return dic


#def dic_keyring( dic, keyring):
def dic_keyring( keyring):
    """ materialized keyring """
    kr_label = keyring.get_label()
    #dic[kr_label] = {}
    dic = {}
    dic["keyring_label"] = kr_label
    dic["keys" ] = []
    for key in keyring.get_all_keys():
        #dic_key(dic[kr_label], key)
        dic["keys" ].append(dic_key(key))
    return dic

def dic_all_keyring():
    """ returns the dictionary representation of all keyrings """

    dic = {}
    dic["keyrings"] = []
    conn = connection_open()
    for keyring in get_all_keyrings(conn):
        if keyring.is_locked():
            keyring.unlock()
        if not keyring.is_locked() and len(keyring.get_label()) > 0:
            #dic_keyring( kr_dic, keyring )
            dic["keyrings"].append(dic_keyring(keyring))
    return dic


from .kr_password_entry import show_password_entry_window


# Exception classes:
class PassException(Exception):
    """All exceptions derive from this class."""
    pass

class NotAvailableException(PassException):
    pass

class LockedException(PassException):
    pass

class KeyNotFoundException(PassException):
    pass

class PromptDismissedException(PassException):
    pass



class Path():
    """ hold the keyringlabel and the keylabel """

    def __init__(self, keyringlabel = None, keylabel=None):
        self.keyringlabel = keyringlabel
        self.keylabel = keylabel

    def set_keyringlabel(self, keyringlabel):
        self.keyringlabel = keyringlabel

    def set_keylabel(self, keylabel):
        self.keylabel = keylabel


class Connection():
    """ This class maintains the passhrase for gpg """

    passphrase = "" 
    locked = True
    keyrings = {}
    keyrings["keyrings"] = []

    def add_keyring(self, keyringlabel):
        keyring = {}
        keyring["label"] = keyringlabel
        keyring["key_list"] = []
        Connection.keyrings["keyrings"].append(keyring)

    def get_keyring(self, keyringlabel):
        for keyring in Connection.keyrings["keyrings"]:
            if keyring["label"] == keyringlabel:
                return keyring
        print("get_keyting: -- not found" )
        return None


    def get_key(self, keyringlabel, keylabel):
        keyring = self.get_keyring(keyringlabel)
        if keyring:
            for key in keyring["key_list"]:
                if key["label"] == keylabel:
                    return key
                else:
                    print("get_key - key not found")
        else:
            print("get_key = keyring not found")
        return None

    def add_key(self, path):

        keyringlabel = path.keyringlabel
        keylabel = path.keylabel
        """ path actually contains the keyring label """
        keyring = self.get_keyring(keyringlabel)
        if keyring:
            key = {}
            key["label"] = keylabel
            key["created"] = key["modified"] = datetime.now()
            key["secret"] = ""
            key["attributes"] = {}
            keyring["key_list"].append(key)
            return
        print( "Cannot add key", keylabel,
                "to keyring", keyringlabel,
               ": keyring not found" )

    def delete_key(self, keyringlabel, keylabel):
        keyring = self.get_keyring(keyringlabel)
        if keyring:
            key = self.get_key(keyringlabel, keylabel)
            if key:
                keyring["key_list"].remove(key)
            else:
                print("key not found")
        else:
            print("keyring not found")

    def delete_keyring(self, keyringlabel):
        keyring = self.get_keyring(keyringlabel)
        self.keyrings["keyrings"].remove(keyring)


    def get_passphrase(self) -> str:
        """ performs validity check and returns the password str,
            or "" when it was cleared """
        return Connection.passphrase

    def set_passphrase(self, passphrase: str) -> None:
        """ set the passphrase for the connection,
        in fact unlocking the connection """
        Connection.passphrase = passphrase
        Connection.locked = False

    def lock(self) -> None:
        """ Locks a connection by clearing the passphrase """
        Connection.passphrase = ""
        Connection.locked = True

    def is_locked(self) -> bool:
        """ Returns the current lock state of a connection """
        return Connection.locked

    def close(self) -> None:
        """ closes the connection, no action needed actually """


def connection_open() -> Connection:
    """ Opens the connection to the pass store, verifies it existence """
    conn = Connection()
    return conn


class Key():
    """Represents a secret item."""
    def __init__(self, connection, path: str, session = None) -> None:
        self.connection = connection
        self.path = path
        connection.add_key(self.path)

    def __eq__(self, other) -> bool:
        return (self.path == other.path)

    def is_locked(self) -> bool:
        """Returns: True if item is locked, otherwise False."""
        return self.connection.is_locked()

    def ensure_not_locked(self) -> None:
        """If kehyring is locked, raises: LockedException"""
        if self.connection.is_locked():
            raise LockedException("Item is locked!")

    def lock(self) -> None:
        """ Lock the connection """
        self.connection.lock()

    def unlock(self) -> bool:
        """ to simulate unlocking, a password prompt is showm,
            the password is captured, and with the password we try to decrypt
            the key. Return True if successful, False otherwise """

        while self.is_locked():
            passwd = show_password_entry_window()
            #if len( passwd ) > 0:
            if passwd != None:
                self.connection.set_passphrase(passwd)
                self.connection.unlock()
                return True
            else:
                return False           # no entry or escape pressed

    def get_key(self):
        key = self.connection.get_key(self.path.keyringlabel,
                                      self.path.keylabel)
        if key:
            return key
        else:
            raise KeyNotFoundException("Key " + self.keylabel + " not found" )

    def get_attributes(self) -> Dict[str, str]:
        """Returns item attributes (dictionary)."""
        key = self.get_key()
        return key["attributes"]

    def set_attributes(self, attributes: Dict[str, str]) -> None:
        """ Sets item attributes to attributes """
        key = self.get_key()
        key["attributes"] = attributes

    def get_label(self) -> str:
        """ retrieves the label for a key """
        return self.path.keylabel

    def set_label(self, label: str) -> None:
        """ the label is the last part of the itempath,
        changing the label means renamiing the file """
        key = self.get_key()
        key["label"] = label
        self.path.set_keylabel(label)

    def delete(self) -> None:
        """ just delete the file? """
        self.connection.delete_key(self.path.keyringlabel,
                                   self.path.keylabel)

    def get_secret(self) -> bytes:
        """Returns item secret (bytestring)."""
        key = self.get_key()
        return key["secret"]

    def get_secret_content_type(self) -> str:
        """ Not supported as such, therefore always return text/plain """
        return "text/plain"

    def set_secret(self, secret: bytes,
                   content_type: str = 'text/plain') -> None:
        """Sets secret to `secret`,
           content_type is there for compat reasons, but is ignored """
        key = self.get_key()
        key["secret"] = secret

    def get_created(self) -> int:
        """Returns UNIX timestamp (integer), when the item was created. """
        key = self.get_key()
        return key["created"]

    def get_modified(self) -> int:
        """Returns UNIX timestamp, when the item was last modified."""
        key = self.get_key()
        return key["modified"]

    def move(self, new_keyring_label):
        key = self.get_key()
        self.connection.delete_key(self.path.keyringlabel,
                                   self.path.keylabel)
        newkeyring = self.connection.get_keyring(new_keyring_label)
        self.path.set_keyringlabel(new_keyring_label)
        newkeyring["key_list"].append(key)


class Keyring():
    """Represents a kehyring."""

    def __init__(self,
                 connection: Connection,
                 path: str = DEFAULT_COLLECTION,
                 session = None) -> None:
        """ the path is the keyring label """
        self.path = Path(keyringlabel = path)
        self.connection = connection
        #connection.add_keyring(self.path.keyringlabel)

    def is_locked(self) -> bool:
        """Returns :const:`True` if item is locked,
           otherwise :const:`False`."""
        return self.connection.is_locked()

    def ensure_not_locked(self) -> None:
        """If keyring is locked, raises LockedException"""
        if self.is_locked():
            raise LockedException('Keyring is locked!')

    def unlock(self) -> bool:
        """ Attempts to unlock a Keyring, 
            currently not using the gpg-agent """
        is_locked = self.is_locked()
        while is_locked:
            passwd = show_password_entry_window()
            #if len( passwd ) > 0:
            if passwd != None:
                self.connection.set_passphrase(passwd)
                is_locked = False
            else:
                break
        return self.is_locked()

    def lock(self) -> None:
        """Locks the keyring."""
        self.connection.lock()

    def delete(self) -> None:
        """Deletes the keyringi and all keys attached to it."""
        self.connection.delete_keyring(self.path.keyringlabel)
        self.keyringlabel = None

    def get_all_keys(self) -> Iterator[Key]:
        """Returns a generator of all keys on the keyring."""
        keyring = self.connection.get_keyring(self.path.keyringlabel)
        if keyring:
            keygen = (Key(self.connection, 
                          Path(self.path.keyringlabel, key["label"])) 
                          for key in keyring["key_list"])
            return keygen
        return

    def search_items(self, attributes: Dict[str, str]) -> Iterator[Key]:
        """Returns a generator of keys with the given attributes."""
        keyring = self.connection.get_keyring(self.path.keyringlabel)
        if keyring:
            for key in keyring["key_list"]:
                key_attributes = key.get_attributes()
                for attrib in key_attributes:
                    if attrib == attributes:
                        yield key
        return


    def get_label(self) -> str:
        """Returns the keyring label."""
        return self.path.keyringlabel

    def set_label(self, label: str) -> None:
        """ the label is the last part of the keyring path,
            changing the label means renamiing the dir """
        self.ensure_not_locked()
        keyring = self.connection.get_keyring(self.path.keyringlabel)
        if keyring:
            self.keyringlabel = keyring["keyring_label"] = label
            self.path.set_keyringlabel(label)


    def create_key(self, label: str, attributes: Dict[str, str],
                   secret: bytes, replace: bool = False,
                   content_type: str = 'text/plain') -> Key:

        """ Creates a key and returns it """
        keypath = Path(keyringlabel = self.path.keyringlabel,
                       keylabel = label)
        key = Key(self.connection, keypath)
        if key:
            key.set_secret(secret)
            key.set_attributes(attributes)
            return key
        return None                     # FIXME: raise an exception ?



def create_keyring(connection: Connection, label: str,
                   alias: str = '', session = None) -> Keyring:
    """Creates a new keyring and returns it. Alias and session are ignored
    for this implementation it only requires a directory to be created """

    keyring = Keyring(connection, label)
    connection.add_keyring(label)
    return keyring


def get_all_keyrings(connection) -> Iterator[Keyring]:
    """Returns a generator of all available keyrings."""

    """
    for con_keyring in connection.keyrings["keyrings"]:
        keyringlabel = con_keyring["label"]
        keyring = Keyring(connection, keyringlabel)
        yield keyring
    """
    keyring_gen = (Keyring(connection, con_keyring["label"])
                   for con_keyring in connection.keyrings["keyrings"])
    return keyring_gen


def get_default_keyring(connection,session = None) -> Keyring:
    """Returns the default keyring. If it doesn't exist, creates it."""
    for keyring in self.get_all_keyrings( connection ):
        if keyring.get_label() == "default":
            return keyring
    return Keyring(connection, "default")


def get_any_keyring(connection) -> Keyring:
    """Returns any keyring, in the following order of preference:
    - The default keyring;
    - The "session" keyring (usually temporary);
    - The first keyring in the keyring list."""

    return get_default_keyring()


def get_keyring_by_alias(connection, alias: str) -> Keyring:
    """Returns the keyring with the given `alias`. If there is no
    such keyring, raises KeyNotFoundException"""

    for keyring in get_all_keyrings(connection):
        if keyring.get_label() == alias:
            return keyring
    raise KeyNotFoundException('No keyring found.')


def search_keys(connection, attributes: Dict[str, str]) -> Iterator[Key]:
    """Returns a generator of keys in all keyrings with the given
    attributes. `attributes` should be a dictionary."""
    for keyring in get_all_keyrings(connection):
        return keyring.search_items(attributes)
    return



#!/usr/bin/env python3

"""
kr_pass: the krapplet pass storage provider
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

import os
import sys
from typing import Dict, Iterator
from subprocess import check_output, STDOUT


# python-gnupg is a Python wrapper for gpg
try:
    import gnupg
except:
    sys.stderr.write("Please install python-gnupg")
    sys.exit(1)

"""
# Assume gtk availability check done in krapplet.py
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
"""

from .kr_password_entry import show_password_entry_window

HOME_DIR = os.path.expanduser("~")
DEFAULT_PASSWORD_STORE = HOME_DIR + "/.password-store"
GPG_ID_FILE = DEFAULT_PASSWORD_STORE + "/.gpg-id"
PASSWORD_STORE = DEFAULT_PASSWORD_STORE
VERBOSE = False
ARMOR = False                         # default no ASCII encoding of secrets
DIR_PERMISSIONS = 0o700               # read/write/execute  user only
FILE_PERMISSIONS = 0o600              # read/write user only
DEFAULT_COLLECTION = "????"
GPG_ID = ""


# Exception classes:
class PassException(Exception):
    """All exceptions derive from this class."""
    pass

class NotAvailableException(PassException):
    pass

class LockedException(PassException):
    pass

class KeyNotFoundException(PassException):
    pass

class PromptDismissedException(PassException):
    pass


def gpg_id():
    """ retrieves the gpg-id from file ${HOME}/.password-store/.gpg-id """
    rval = ""
    try:
        with open(GPG_ID_FILE) as file_handle:
            rval = file_handle.read()
    except:
        print("Gpg_id not found")
    return rval


def encrypt(unenc: str, recipients: str, passphrase: str) -> str:
    """ Encrypts a string using the passphrase """
    gpg = gnupg.GPG(verbose=VERBOSE, use_agent=False)
    enc = gpg.encrypt(unenc, passphrase=passphrase, sign=True,
                      recipients = recipients, armor = ARMOR)
    if enc.ok:
        return enc.data
    return ""


def passwd_test( recipients: str, passphrase: str) -> bool:
    """ tests if we can encrypt a string """
    enc = encrypt("An unecrypted string", recipients, passphrase)
    return len( enc ) > 0


class Connection():
    """ This class maintains the passhrase for gpg """

    passphrase = ""                             # class var over all instances
    recipients = gpg_id().strip()

    def check_validty(self) -> None:
        """ Clears the password when encryption of a string fails """
        return
        if not passwd_test(Connection.recipients, Connection.passphrase):
            self.lock()

    def get_passphrase(self) -> str:
        """ performs validity check and returns the password str,
            or "" when it was cleared """
        self.check_validty()
        return Connection.passphrase

    def set_passphrase(self, passphrase: str) -> None:
        """ set the passphrase for the connection,
        in fact unlocking the connection """
        if passwd_test(Connection.recipients, Connection.passphrase):
            Connection.passphrase = passphrase

    def lock(self) -> None:
        """ Locks a connection by clearing the passphrase """
        Connection.passphrase = ""
        check_output(["gpgconf", "--reload", "gpg-agent"],
                      stderr=STDOUT ).decode("utf-8")

    def is_locked(self) -> bool:
        """ Returns the current lock state of a connection """
        self.check_validty()
        is_locked = Connection.passphrase == ""
        return is_locked

    def close(self) -> None:
        """ closes the connection, no action needed actually """


def connection_open() -> Connection:
    """ Opens the connection to the pass store, verifies it existence """
    conn = Connection()
    return conn


def set_armor(val: bool) -> None:
    """ Sets the global variabe ARMOR """
    global ARMOR
    ARMOR = val


def check_gpg( gpg_id ) -> None:
    """ rudimentary check for a gpg_id, exits when gpg
    command or gpg-id not found """
    return

"""
class PasswordEntryDialog(Gtk.Dialog):
    "" PasswordEntryDialog: pops up a window to ask for a password ""
    def __init__(self, parent, title: str, unlock_item: str) -> None:
        Gtk.Dialog.__init__(self, title=title, flags=0)
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.set_default_size(150, 100)
        box = self.get_content_area()
        password_prompt = Gtk.Label(label=unlock_item, xalign=0)
        self.passord_entry = Gtk.Entry(xalign = 0, visibility=False)
        self.passord_entry.set_activates_default(True)
        self.set_default_response(Gtk.ResponseType.OK)
        grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        box.add(grid)
        grid.attach_next_to(password_prompt, None,
                            Gtk.PositionType.BOTTOM, 1, 1)
        grid.attach_next_to(self.passord_entry, password_prompt,
                            Gtk.PositionType.RIGHT, 1, 1)
        self.show_all()

    def get_pw( self ) -> str:
        "" returns the entered passwd from screen ""
        return self.passord_entry.get_text()
"""

class Key():
    """Represents a secret item."""
    def __init__(self, connection, path: str, session = None) -> None:
        self.path = path
        self.session = session
        self.connection = connection
        self.attrs = {}
        self.secret = ""
        self.recipients = gpg_id().strip()
        self.created = self.modified = 0

    def __eq__(self, other) -> bool:
        assert isinstance(other.path, str)
        return self.path == other.path

    def save_encrypted(self) -> None:
        """ saves key data encrypted"""
        unenc_str = self.secret + "\n"
        for name in self.attrs.keys():
            unenc_str += name + ": " + self.attrs[name] + "\n"
        enc = encrypt(unenc_str, self.connection.recipients,
                      self.connection.get_passphrase())
        if len(enc) > 0:
            with open(os.open(self.path,
                              os.O_CREAT | os.O_WRONLY, FILE_PERMISSIONS),
                      "wb") as file_handle:
                file_handle.write( enc )

    def read_encrypted(self) -> bool:
        """ reads the key from file """
        gpg = gnupg.GPG(verbose=VERBOSE, use_agent=False,
                        options="--ignore-mdc-error")
        passphrase=self.connection.get_passphrase()
        with open(self.path, 'rb') as file_handle:
            dec = gpg.decrypt_file(file_handle, passphrase=passphrase)
        if dec.ok:
            #_, fname = os.path.split(self.path)   # don't think this needed
            #label, _ = os.path.splitext(fname)
            self.created = self.modified = 0
            lines = str(dec).splitlines()
            secret = ""
            for line in lines:
                if len(secret) == 0:           # first line contains the secret
                    secret = line
                    self.secret = secret
                else:                          # next lines contain the attribs
                    colon_idx = line.find(':')
                    name = line[:colon_idx]
                    val = line[colon_idx+2:]
                    self.attrs[name] = val
            return True
        print("Could not read decrypted file, status:", dec.status)
        return False

    def is_locked(self) -> bool:
        """Returns :const:`True` if item is locked, otherwise :const:`False`."""
        return self.connection.is_locked()

    def ensure_not_locked(self) -> None:
        """If kehyring is locked, raises: LockedException"""
        if self.connection.is_locked():
            self.unlock()
            #raise LockedException("Item is locked!')

    def lock(self) -> None:
        """ Lock the connection """
        self.connection.lock()

    def unlock(self) -> bool:
        """ to simulate unlocking, a password prompt is showm,
            the password is captured, and with the password we try to decrypt
            the key. Return True if successful, False otherwise """

        while self.is_locked():
            passwd = show_password_entry_window()
            #if len( passwd ) > 0:
            if passwd != None:
                self.connection.set_passphrase(passwd)
                if self.read_encrypted():
                    return True
                self.lock()
            else:
                return False           # no entry or escape pressed

    def get_attributes(self) -> Dict[str, str]:
        """Returns item attributes (dictionary)."""
        if self.attrs == {}:
            self.read_encrypted()
        return self.attrs

    def set_attributes(self, attributes: Dict[str, str]) -> None:
        """ Sets item attributes to attributes """
        self.ensure_not_locked()
        self.attrs = attributes
        self.save_encrypted()

    def get_label(self) -> str:
        """ retrieves the label for a key """
        _, fname = os.path.split(self.path)
        label, _ = os.path.splitext(fname )
        return label

    def set_label(self, label: str) -> None:
        """ the label is the last part of the itempath,
        changing the label means renamiing the file """
        self.ensure_not_locked()
        dname, fname = os.path.split(self.path)
        _, ext = os.path.splitext(fname)
        new_fname = label + ext
        newpath = os.path.join(dname, new_fname)
        if not self.path == newpath:
            os.rename(self.path, newpath)
            self.path = newpath

    def delete(self) -> None:
        """ just delete the file? """
        os.remove(self.path)
        self.attrs = {}
        self.secret = ""

    def get_secret(self) -> bytes:
        """Returns item secret (bytestring)."""
        if len( self.secret ) == 0:
            self.read_encrypted()
        return str.encode(self.secret)

    def get_secret_content_type(self) -> str:
        """ Not supported as such, therefore always return text/plain """
        return "text/plain"

    def set_secret(self, secret: bytes,
                   content_type: str = 'text/plain') -> None:
        """Sets secret to `secret`,
           content_type is there for compat reasons, but is ignored """
        self.ensure_not_locked()
        self.secret = secret.decode("utf-8")
        self.save_encrypted()

    def get_created(self) -> int:
        """Returns UNIX timestamp (integer), when the item was created. """
        statinfo = os.stat(self.path)
        return statinfo.st_ctime

    def get_modified(self) -> int:
        """Returns UNIX timestamp (integer), when the item was last modified."""
        statinfo = os.stat(self.path)
        return statinfo.st_mtime

    def move(self, new_keyring_label):
        """ moves a key to a different keyring """
        keyrings = get_all_keyrings(self.connection)
        for keyring in keyrings:
            if keyring.get_label() == new_keyring_label:
                _, fname = os.path.split(self.path)
                newpath = os.path.join(keyring.path, fname)
                os.replace(self.path, newpath)
                self.path = newpath
                return


class Keyring():
    """Represents a kehyring."""

    def __init__(self,
                 connection: Connection,
                 path: str = DEFAULT_COLLECTION,
                 session = None) -> None:
        self.connection = connection
        self.session = session
        self.path = path

    def is_locked(self) -> bool:
        """Returns :const:`True` if item is locked,
           otherwise :const:`False`."""
        is_locked = self.connection.is_locked()
        return is_locked

    def ensure_not_locked(self) -> None:
        """If keyring is locked, raises LockedException"""
        if self.is_locked():
            raise LockedException('Keyring is locked!')

    def unlock(self) -> bool:
        """ Attempts to unlock a Keyring, 
            currently not using the gpg-agent """
        is_locked = self.is_locked()
        while is_locked:
            passwd = show_password_entry_window()
            #if len( passwd ) > 0:
            if passwd != None:
                if passwd_test(self.connection.recipients, passwd):
                    self.connection.set_passphrase(passwd)
                is_locked = self.is_locked()
            else:
                break
        return self.is_locked()

    def lock(self) -> None:
        """Locks the keyring."""
        self.connection.lock()

    def delete(self) -> None:
        """Deletes the keyringi and all keys attached to it."""
        self.ensure_not_locked()
        for dirname, dirs, files in os.walk(self.path, topdown = False):
            for directory in dirs:
                os.rmdir(os.path.join(dirname, directory))
            for fil in files:
                os.remove(os.path.join(dirname, fil))
        os.rmdir(self.path)
        self.path = None

    def get_all_keys(self) -> Iterator[Key]:
        """Returns a generator of all keys on the keyring."""
        for fname in os.listdir(self.path):
            keypath = os.path.join(self.path, fname)
            if os.path.isfile(keypath) and keypath.endswith(".gpg"):
                yield Key(self.connection, keypath, None)

    def search_items(self, attributes: Dict[str, str]) -> Iterator[Key]:
        """Returns a generator of keys with the given attributes."""
        if not self.is_locked():
            for key in self.get_all_keys():
                key_attributes = key.get_attributes()
                for attrib in key_attributes:
                    if attrib == attributes:
                        return key
        else:
            return
            yield

    def get_label(self) -> str:
        """Returns the keyring label."""
        _, label = os.path.split( self.path )
        return label

    def set_label(self, label: str) -> None:
        """ the label is the last part of the keyring path,
            changing the label means renamiing the dir """
        self.ensure_not_locked()
        path, _ = os.path.split(self.path)
        newpath = os.path.join(path, label)
        if not self.path == newpath:
            os.rename(self.path, newpath)
            self.path = newpath


    def create_key(self, label: str, attributes: Dict[str, str],
                   secret: bytes, replace: bool = False,
                   content_type: str = 'text/plain') -> Key:

        """ Creates a key and returns it """
        self.ensure_not_locked()
        path = os.path.join(self.path, label + ".gpg")
        if replace or not os.path.isfile(path):
            key = Key(self.connection, path)
            key.set_secret(secret)
            key.set_attributes(attributes)
            return key
        return None                     # FIXME: raise an exception ?



def create_keyring(connection: Connection, label: str,
                   alias: str = '', session = None) -> Keyring:
    """Creates a new keyring and returns it. Alias and session are ignored
    for this implementation it only requires a directory to be created """

    path =  os.path.join(DEFAULT_PASSWORD_STORE, label)
    if not os.path.exists(path):
        os.mkdir(path)
    if os.path.isdir(path):
        os.chmod(path, DIR_PERMISSIONS)
        return Keyring(connection, path)
    return None


def get_all_keyrings(connection) -> Iterator[Keyring]:
    """Returns a generator of all available keyrings."""
    for dname in os.listdir(DEFAULT_PASSWORD_STORE):
        path = os.path.join(DEFAULT_PASSWORD_STORE, dname)
        if os.path.isdir(path):
            yield Keyring(connection, path)


def get_default_keyring(connection,session = None) -> Keyring:
    """Returns the default keyring. If it doesn't exist, creates it."""
    defaultpath = os.path.join(DEFAULT_PASSWORD_STORE, "default")
    if os.path.isdir(defaultpath):
        return Keyring(connection, defaultpath)
    return Keyring( connection, "default" )

def get_any_keyring(connection) -> Keyring:
    """Returns any keyring, in the following order of preference:
    - The default keyring;
    - The "session" keyring (usually temporary);
    - The first keyring in the keyring list."""

    defaultpath = os.path.join(DEFAULT_PASSWORD_STORE, "default")
    if os.path.isdir(defaultpath):
        return Keyring(connection, defaultpath)
    for entry in os.listdir(DEFAULT_PASSWORD_STORE):
        path = path.join(DEFAULT_PASSWORD_STORE, entry)
        if os.path.isdir(path):
            return Keyring(connection, path)
    raise KeyNotFoundException('No keyring found.')


def get_keyring_by_alias(connection, alias: str) -> Keyring:
    """Returns the keyring with the given `alias`. If there is no
    such keyring, raises KeyNotFoundException"""

    path = os.path.join(DEFAULT_PASSWORD_STORE, alias)
    if os.path.isdir(path):
        return Keyring(connection, path)
    raise KeyNotFoundException('No keyring found.')


def search_keys(connection, attributes: Dict[str, str]) -> Iterator[Key]:
    """Returns a generator of keys in all keyrings with the given
    attributes. `attributes` should be a dictionary."""
    for entry in os.listdir(DEFAULT_PASSWORD_STORE):
        path = os.path.join(DEFAULT_PASSWORD_STORE, entry)
        if os.path.isdir(path):
            keyring = Keyring(connection, path)
            return keyring.search_items(attributes)
    return



#!/usr/bin/env python3

"""
kr_test: the krapplet test suite
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.

Usage: python -m unittest krapplet.kr_test

"""



import unittest
import secrets

#import krapplet.kr_pass
from .kr_storage_select import StorageSelect
st_sel=StorageSelect("pass")

from . import kr_json
from . import kr_pass
from . import kr_crypt
from . import kr_memstorage
from .krapplet import gen_pw


#class TestStringMethods(unittest.TestCase):
class test_kr_passs(unittest.TestCase):
    """ kr_pass tests: for the pass compatible storage provider """

    def setUp(self):
        """ test the creation/deletion of keyrings and keys """
        self.test_passwd = "__test_keyring_pw"
        self.keyring_name = "__test_keyring"
        self.keyname = "__test_key"
        self.attrs = { "URL": "http://www.xyz.com" }
        self.secret = bytes("secret", 'utf-8')
        self.conn = kr_pass.connection_open()
        self.conn.set_passphrase(self.test_passwd)
        self.keyring = kr_pass.create_keyring(self.conn,
                                                       self.keyring_name)
        if self.keyring.is_locked():
            self.keyring.unlock()
        self.key = self.keyring.create_key(self.keyname,
                                           self.attrs,
                                           self.secret)

    def tearDown(self):
        """ deletes the test key and test keyring """
        self.key.delete()
        self.keyring.delete()
        self.conn.close()

    def test_pass_key(self):
        """ Finds the key and sees if it has the right value """
        conn = kr_pass.connection_open()
        conn.set_passphrase(self.test_passwd)
        test_success = False
        for keyring in kr_pass.get_all_keyrings(conn):
            if keyring.get_label() == self.keyring_name:
                for key in keyring.get_all_keys():
                    if key.get_label() == self.keyname:
                        self.assertEqual(key.get_attributes(), self.attrs)
                        self.assertEqual(key.get_secret(), self.secret)
                        test_success = True
        conn.close()
        self.assertTrue(test_success)

    def test_update_key(self):
        """ tests updating a key's attrs, and secret """
        conn = kr_pass.connection_open()
        conn.set_passphrase(self.test_passwd)
        attrs = self.attrs
        for key in kr_pass.search_keys(conn, attrs):
            self.assertEqual(key.get_attributes(), self.attrs)
            new_attrs = { "Num": "001" }
            key.set_attributes(attrs)
            self.assertEqual(key.get_attributes(), attrs)
            new_label = "NewLabel"
            key.set_label(new_label)
            self.assertEqual(key.get_label(), new_label)
            new_secret = bytes("NewSecret", 'utf-8')
            key.set_secret(new_secret)
            self.assertEqual(key.get_secret(), new_secret)
        conn.close()


"""
class test_kr_memstorage(unittest.TestCase):
    def setUp(self):
        self.con = kr_memstorage.connection_open()

    def test_1(self):
        self.assertTrue(self.con.is_locked())
        krlabel = "kr_label"
        k_label = "k_label"
        k_attr = {}
        k_attr["type"] = "loper"
        k_attr["color"] = "grey"
        k_secret = "5ecr3T"
        kr = kr_memstorage.create_keyring(self.con, krlabel)
        self.assertEqual(kr.get_label(), krlabel)
        k =  kr.create_key(k_label, k_attr, k_secret)
        self.assertEqual(k.get_label(), k_label)
        self.assertEqual(k.get_attributes(), k_attr)
        self.assertEqual(k.get_secret(), k_secret)
        k_newlabel = "k_newlabel"
        k_newattr = {}
        k_newattr["newtype"] = "newloper"
        k_newattr["newtype"] = "newloper"
        k_newsecret = "new5ecr3T"
        k.set_label(k_newlabel)
        self.assertEqual(k.get_label(), k_newlabel)
        k.set_attributes(k_newattr)
        self.assertEqual(k.get_attributes(), k_newattr)
        k.set_secret(k_newsecret)
        self.assertEqual(k.get_secret(), k_newsecret)
        # something with created and modified times..? 
        krnewlabel = "kr2_label"
        kr2 = kr_memstorage.create_keyring(self.con, krnewlabel)
        k.move(krnewlabel)
        print( "key delete" )
        k.delete()
        print( "keyring delete" )
        kr.delete()
        print( "keyring2 delete" )
        kr2.delete()

"""

class test_kr_json(unittest.TestCase):

    def test_json(self):
        """ Test of the json module """
        kr_dic = kr_json.dic_all_keyring()
        kr_str = kr_json.dic2str(kr_dic)
        kr_dic2 = kr_json.str2dic(kr_str)
        self.assertEqual(kr_dic, kr_dic2)
        kr_str2 = kr_json.dic2str(kr_dic2)
        self.assertEqual(kr_str, kr_str2)



class test_kr_crypt(unittest.TestCase):
    "Test symetric  encryption"
    def setUp(self):
        self.secret_msg = "!terces siht wonk ot si eno oN"
        self.passphrase = "S3creT---!"

    def test_crypt(self):
        " test encryption "
        encrypted_secret_msg = kr_crypt.sym_encrypt(
            self.secret_msg, self.passphrase)
        decrypted_secret_msg = kr_crypt.sym_decrypt(
            encrypted_secret_msg, self.passphrase).decode("utf-8")
        self.assertEqual(self.secret_msg, decrypted_secret_msg)
        """
        wrong_passphrase = "S3creT---?"
        failed_decrypted_secret_msg = kr_crypt.sym_decrypt(
            encrypted_secret_msg, wrong_passphrase)
        self.assertIsNone(failed_decrypted_secret_msg)
        """


class test_kr_asymcrypt(unittest.TestCase):
    "Test for asymetric encryption"
    def setUp(self):
        self.secret_msg = "!terces siht wonk ot si eno oN"
        self.passphrase = b"S3creT---!"
        self.alice_privkey = kr_crypt.asym_genkey()
        self.alice_pubkey = kr_crypt.asym_pubkey(self.alice_privkey)
        self.bob_privkey = kr_crypt.asym_genkey()
        self.bob_pubkey = kr_crypt.asym_pubkey(self.bob_privkey)
        self.salt = secrets.token_bytes()

    def test_asym_shared_keys(self):
        "Test diffie-hellman key exchange"
        alice_sh_key = kr_crypt.asym_derived_sharedkey(
            self.alice_privkey, self.bob_pubkey, self.salt, "test")
        bob_sh_key = kr_crypt.asym_derived_sharedkey(
            self.bob_privkey, self.alice_pubkey, self.salt, "test")

        # Alice and Bob's shared keys should be the same:
        self.assertEqual(alice_sh_key, bob_sh_key)

        # wrong shared key uses the wrong salt:
        wrong_sh_key = kr_crypt.asym_derived_sharedkey(
            self.bob_privkey, self.alice_pubkey, self.salt + b'ss', "tEst")
        self.assertNotEqual(alice_sh_key, wrong_sh_key)

    def test_privkey_serialization_no_passwd(self):
        "Test for serializing a private key with no passwd"
        ser_privkey = kr_crypt.asym_serialize_privkey(
            self.alice_privkey, None)
        privkey = kr_crypt.asym_deserialize_privkey(ser_privkey, None)
        self.assertIsNotNone(privkey)
        if privkey:
            ser2_privkey = kr_crypt.asym_serialize_privkey( privkey, None)
            self.assertEqual(ser_privkey, ser2_privkey)

    def test_privkey_serialization_with_passwd(self):
        "Test for serializing a private key with a passwd"
        ser_privkey = kr_crypt.asym_serialize_privkey(
            self.alice_privkey, self.passphrase)
        privkey = kr_crypt.asym_deserialize_privkey(ser_privkey,
                                                    self.passphrase)
        self.assertIsNotNone(privkey)
        """ 
        if privkey:
            ser2_privkey = kr_crypt.asym_serialize_privkey(privkey,
                                                           self.passphrase)
            self.assertEqual(ser_privkey, ser2_privkey) """


    def test_privkey_serialization_with_passwd_fail(self):
        "Test for serializing a private key with a passwd"
        ser_privkey = kr_crypt.asym_serialize_privkey(
            self.alice_privkey, self.passphrase)
        privkey = kr_crypt.asym_deserialize_privkey(ser_privkey, "a_guess")
        self.assertIsNone(privkey)

    def test_asym_signing(self):
        "Test signing verification"
        msg2sign = b'This is a very important message'
        alice_sign = kr_crypt.asym_signature(
            msg2sign, self.alice_privkey)
        self.assertTrue(kr_crypt.asym_signature_verify(
            msg2sign, alice_sign, self.alice_pubkey))

    def test_asym_signing_fail(self):
        "Test signing verification"
        msg2sign = b'This is a very important message'
        alice_sign = kr_crypt.asym_signature(
            msg2sign, self.alice_privkey)
        bad_sign = (10,10)
        self.assertFalse(kr_crypt.asym_signature_verify(
            msg2sign, bad_sign, self.alice_pubkey))





if __name__ == '__main__':
    unittest.main()





#!/usr/bin/env python

"""
kr_password_entry: facilittes password entry from screen
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

# Assume gtk availability check done in krapplet.py
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

# shortcuts so that calling code does not need to import Gtk
OPEN = Gtk.FileChooserAction.OPEN
SAVE = Gtk.FileChooserAction.SAVE
SELECT_FOLDER = Gtk.FileChooserAction.SELECT_FOLDER
CREATE_FOLDER = Gtk.FileChooserAction.CREATE_FOLDER

class FilePicker():

    def __init__(self, title, fname, action):
        #dialog = Gtk.FileChooserDialog(title=title,
        #                               parent=self, action=action)
        dialog = Gtk.FileChooserDialog(title=title,action=action)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN, Gtk.ResponseType.OK,)
        dialog.set_current_name(fname)
        self.filename = None
        #self.add_filters(dialog)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.filename = dialog.get_filename()
        #elif response == Gtk.ResponseType.CANCEL:
        dialog.destroy()

    def selected_file(self):
        return self.filename

    def add_filters(self, dialog):
        """
        filter_text = Gtk.FileFilter()
        filter_text.set_name("Text files")
        filter_text.add_mime_type("text/plain")
        dialog.add_filter(filter_text)
:q
        filter_py = Gtk.FileFilter()
        filter_py.set_name("Python files")
        filter_py.add_mime_type("text/x-python")
        dialog.add_filter(filter_py)

        filter_any = Gtk.FileFilter()
        filter_any.set_name("Any files")
        filter_any.add_pattern("*")
        dialog.add_filter(filter_any) """




#!/usr/bin/python3
"""
krapplet: A password manager written as a gnome-keyring applet
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies
"""

from . import krapplet

krapplet.main()



#!/usr/bin/env python

"""
kr_password_entry: facilittes password entry from screen
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

# Assume gtk availability check done in krapplet.py
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class PasswordEntryDialog(Gtk.Dialog):
    """ PasswordEntryDialog: pops up a window to ask for a password """
    def __init__(self, title: str, prompt: str) -> None:
        Gtk.Dialog.__init__(self, title=title, flags=0)
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.set_default_size(150, 100)
        box = self.get_content_area()
        password_prompt = Gtk.Label(label=prompt, xalign=0)
        self.passord_entry = Gtk.Entry(xalign = 0, visibility=False)
        self.passord_entry.set_activates_default(True)
        self.set_default_response(Gtk.ResponseType.OK)
        grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        box.add(grid)
        grid.attach_next_to(password_prompt, None,
                            Gtk.PositionType.BOTTOM, 1, 1)
        grid.attach_next_to(self.passord_entry, password_prompt,
                            Gtk.PositionType.RIGHT, 1, 1)
        self.show_all()

    def get_pw( self ) -> str:
        """ returns the entered passwd from screen """
        return self.passord_entry.get_text()


def show_password_entry_window(title = "Unlock", prompt = "Passphrase"):
    """ shows the password entry window, return an entered passwd, or
        None when escape was pressed"""

    passwd = ""
    dialog = PasswordEntryDialog(title = title, prompt = "Passphrase")
    response = dialog.run()
    if response == Gtk.ResponseType.OK:
        passwd = dialog.get_pw()
    elif response == Gtk.ResponseType.CANCEL:
        passwd = None
    dialog.destroy()
    return passwd



#!/usr/bin/env python

"""
kr_backup: functions for backing up keyrings and keys
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

import os
from datetime import datetime

# Assume gtk availability check done in krapplet.py
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from . import kr_crypt
from . import kr_json
from . import kr_password_entry
from . import kr_filepicker
from . import kr_storage
from .kr_popup import show_error_message, show_info_message
from .kr_time import timestamp

FRAME_LABEL_XALIGN = 0.025

def backup(path = None):
    """ backups all keysring and keys to a file indicated by path
    if path is None, a file picker option will be provided """

    kr_dic = kr_json.dic_all_keyring()
    #kr_str = json.dumps(kr_dic)
    kr_str = kr_json.dic2str(kr_dic)
    pw_prompt = "Passphrase for backup file"
    passphrase = kr_password_entry.show_password_entry_window(
        title = pw_prompt)
    kr_enc_str = kr_crypt.sym_encrypt(kr_str, passphrase)
    now = datetime.now()
    timestamp_formatstr = "%Y-%m-%d_%H:%M"
    timestamp = now.strftime(timestamp_formatstr)
    backup_fname = "krapplet_" + timestamp + ".bak"
    if path:
        backup_fname = os.path.join(path, backup_fname)
    else:
        folder_picker = kr_filepicker.FilePicker("Backup", backup_fname,
                                                 kr_filepicker.SAVE)
        #folder_picker.dialog.run()
        backup_fname = folder_picker.selected_file()
    if backup_fname:                 # would be None on escape key
        try:
            with open(backup_fname, "wb") as backup_file:
                backup_file.write(kr_enc_str)
                info_msg = "Backup file: " + backup_fname
                show_info_message(None, "Backup successful", info_msg)
        except PermissionError:
            errormsg = "Cannot open file " + backup_fname + " for writing"
            show_error_message(None, "Permission error", errormsg )
        except Exception as exc:
            show_error_message(None, "Error saving backup", str(exc))



class  RestoreDialog(Gtk.Dialog):
    """ class to show keyrings, keys, and attribs, for restore function """
    def __init__(self, backup_fname, dic):
        Gtk.Dialog.__init__(self, title="Restore " + backup_fname, flags=0 )
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        content_area = self.get_content_area()
        self.grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        content_area.add(self.grid)
        self.keyring_frame = Gtk.Frame(label="Keyrings", 
                                       label_xalign=FRAME_LABEL_XALIGN)
        self.grid.attach_next_to(self.keyring_frame, None,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.keyringbox = Gtk.Box(orientation = Gtk.Orientation.VERTICAL,
                                  border_width = 5)
        self.keyring_frame.add(self.keyringbox)
        self.key_frame = self.attr_frame = None
        self.restore_all_keys_button = self.restore_button = None

        keyrings = dic["keyrings"]
        kr_group = None
        for keyring in keyrings:
            kr_button = Gtk.RadioButton.new_with_label_from_widget(
                kr_group, keyring["keyring_label"][:32])
            kr_button.connect("toggled", self.show_keyring, keyring)
            self.keyringbox.pack_start(kr_button, False, False, 0)
            if not kr_group:            # first one
                kr_group = kr_button
                self.show_keyring( kr_button, keyring)
        restore_all_keyrings_button = Gtk.Button("Restore all keyrings",
                                                 border_width = 10)
        restore_all_keyrings_button.connect(
            "clicked", self.restore_all_keyrings, keyrings)
        self.grid.attach_next_to(restore_all_keyrings_button,
                                 self.keyring_frame,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.show_all()

    def clear_keybox(self):
        "clear the keyring box"
        if self.key_frame:
            self.key_frame.destroy()
            self.key_frame = None
        if self.restore_all_keys_button:
            self.restore_all_keys_button.destroy()
            self.restore_all_keys_button = None

    def clear_attrbox(self):
        "clear the keyring box"
        if self.attr_frame:
            self.attr_frame.destroy()
            self.attr_frame = None
        if self.restore_button:
            self.restore_button.destroy()
            self.restore_button = None

    def show_keyring(self, button, keyring):
        """ Shows the keys in a keyring """
        if not button.get_active():       # only when activated
            self.clear_keybox()
            self.clear_attrbox()
            return
        #self.key_frame = Gtk.Frame(label="Keys",
        self.key_frame = Gtk.Frame(
            label=keyring["keyring_label"][:20] + " keys",
            label_xalign=FRAME_LABEL_XALIGN)
        self.grid.attach_next_to(self.key_frame, self.keyring_frame,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.keybox = Gtk.Box(orientation = Gtk.Orientation.VERTICAL,
                              border_width = 5)
        self.key_frame.add(self.keybox)
        key_group = None
        for key in keyring["keys"]:
            keybutton = Gtk.RadioButton.new_with_label_from_widget(
                key_group, key["key_label"][:32])
            keybutton.connect("toggled", self.show_key, keyring, key)
            self.keybox.pack_start(keybutton, False, False, 0)
            if not key_group:
                key_group = keybutton
                self.show_key(keybutton, keyring, key)
        self.restore_all_keys_button = Gtk.Button("Restore all keys",
                                                   border_width = 10)
        self.restore_all_keys_button.connect("clicked",
                                             self.restore_all_keys, keyring)
        self.grid.attach_next_to(self.restore_all_keys_button,
                                 self.key_frame,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.show_all()


    def show_key(self, button, keyring, key):
        if not button.get_active():
            self.clear_attrbox()
            return
        #self.attr_frame = Gtk.Frame(label="Attributes",
        self.attr_frame = Gtk.Frame(
            label=key["key_label"][:32] + " attributes",
            label_xalign=FRAME_LABEL_XALIGN)
        self.grid.attach_next_to(self.attr_frame, self.key_frame,
                                 Gtk.PositionType.RIGHT, 1, 1)
        key_grid = Gtk.Grid(row_spacing=2, column_spacing=5,border_width=5)
        self.attr_frame.add(key_grid)
        created_prompt = Gtk.Label(label="Created", xalign=0)
        created_value = Gtk.Label(label=timestamp(int(key["created"])),
                                  xalign=0)
        key_grid.attach_next_to(created_prompt,None,
                                Gtk.PositionType.BOTTOM, 1, 1)
        key_grid.attach_next_to(created_value, created_prompt,
                                Gtk.PositionType.RIGHT, 1, 1)

        modified_prompt = Gtk.Label(label="Modified", xalign=0)
        modified_value = Gtk.Label(label=timestamp(int(key["modified"])),
                                   xalign=0)
        key_grid.attach_next_to(modified_prompt, created_prompt,
                                Gtk.PositionType.BOTTOM, 1, 1)
        key_grid.attach_next_to(modified_value, modified_prompt,
                                Gtk.PositionType.RIGHT, 1, 1)

        last_prompt = modified_prompt
        attribs = key["attributes"]
        for attr in attribs:
            attr_prompt = Gtk.Label(label=attr[:32], xalign=0)
            attr_value = Gtk.Label(label=attribs[attr][:32], xalign=0)
            key_grid.attach_next_to(attr_prompt, last_prompt,
                                    Gtk.PositionType.BOTTOM, 1, 1)
            key_grid.attach_next_to(attr_value, attr_prompt,
                                    Gtk.PositionType.RIGHT, 1, 1)
            last_prompt = attr_prompt

        secret_pronpt = Gtk.Label(label="Secret", xalign=0)
        masked_secret = key["secret"][:2] + " ... " + key["secret"][-2:]
        secret_value = Gtk.Label(label=masked_secret, xalign=0)
        key_grid.attach_next_to(secret_pronpt, last_prompt,
                                Gtk.PositionType.BOTTOM, 1, 1)
        key_grid.attach_next_to(secret_value, secret_pronpt,
                                Gtk.PositionType.RIGHT, 1, 1)
        self.restore_button = Gtk.Button("Restore Key", border_width = 10)
        self.restore_button.connect("clicked", 
                                    self.restore_key, keyring, key)
        self.grid.attach_next_to(self.restore_button, self.attr_frame,
                                 Gtk.PositionType.BOTTOM, 1, 1)

        self.show_all()
        return

    def restore_all_keyrings(self, button, keyrings):
        " Restores all the keyrings from the backup "
        log = []
        for keyring in keyrings:
            for rkey in keyring["keys"]:
                self.restore_one_key(keyring, rkey, log)
        if len(log) > 0:
            show_info_message(None, "Restore log", log)

    def restore_all_keys(self, button, restore_keyring):
        " Restores all keys in a keyring "
        log = []
        for rkey in restore_keyring["keys"]:
            self.restore_one_key(restore_keyring, rkey, log)
        if len(log) > 0:
            show_info_message(None, "Restore log", log)

    def restore_key(self, button, restore_keyring, restore_key):
        log = []
        self.restore_one_key(restore_keyring, restore_key, log)
        if len(log) > 0:
            show_info_message(None, "Restore log", log)

    def restore_one_key(self, restore_keyring, restore_key, log):
        " restores a key to the main storage "
        update_done = False
        conn = kr_storage.connection_open()
        for keyring in kr_storage.get_all_keyrings(conn):
            kr_label = keyring.get_label()
            if kr_label == restore_keyring["keyring_label"]:
                for key in keyring.get_all_keys():
                    k_label = key.get_label()
                    if k_label == restore_key["key_label"]:
                        key.set_attributes(restore_key["attributes"])
                        key.set_secret(restore_key["secret"].encode())
                        update_done = True
                        log.append("Key "
                                   + k_label
                                   + " in keyring "
                                   + kr_label
                                   + ": attributes and secret updated.")
                if not update_done:       # key not found
                    if keyring.is_locked():
                        keyring.unlock()
                    if not keyring.is_locked():
                        key = keyring.create_key(
                            restore_key["key_label"],
                            restore_key["attributes"],
                            restore_key["secret"].encode())
                        log.append("Key "
                                   + restore_key["key_label"]
                                   + " in keyring "
                                   + kr_label
                                   + ": created.")
                    update_done = True
        if not update_done:               # keyring not found
            keyring = kr_storage.create_keyring(
                conn, restore_keyring["keyring_label"])
            log.append("Keyring "
                       + restore_keyring["keyring_label"]
                       + ": created.")
            if keyring.is_locked():
                keyring.unlock()
            if not keyring.is_locked():
                key = keyring.create_key(restore_key["key_label"],
                                         restore_key["attributes"],
                                         restore_key["secret"].encode())
                log.append("Key "
                           + restore_key["key_label"]
                           + " in keyring "
                           + restore_keyring["keyring_label"]
                           + ": created")

            else:
                show_error_message(None, 
                                   "Keyring not restored", "Keyring locked")
        return


def restore():
    """ restores keyrings and keys """
    backup_fname = ""
    folder_picker = kr_filepicker.FilePicker("Restore", backup_fname,
                                              kr_filepicker.OPEN)
    backup_fname = folder_picker.selected_file()
    enc_data = None
    if backup_fname:
        try: 
            with open(backup_fname, "rb") as backup_file:
                enc_data = backup_file.read()
        except PermissionError:
            errormsg = "Cannot open file " + backup_fname + " for reading"
            show_error_message(None, "Permission error", errormsg )
        except Exception as exc:
            show_error_message(None, "Error reading backup", str(exc))
    if enc_data:
        pw_prompt = "Passphrase for backup file"
        done = False
        while not done:
            passphrase = kr_password_entry.show_password_entry_window(
                title = pw_prompt)
            if passphrase:
                decstr = kr_crypt.sym_decrypt(enc_data, passphrase)
                if decstr:
                    dic = kr_json.str2dic(decstr)
                    dia = RestoreDialog(backup_fname, dic)
                    dia.run()
                    dia.destroy()
                    done = True
            else:
                done = True





#!/usr/bin/env python3
"""
krapplet: A password manager written as a gnome-keyring applet
krapplet.py is the main program
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies
"""

# standard Python modules
import sys
import os
import signal
import string
import secrets
import random
import locale
import webbrowser
import subprocess
import math
import argparse
from datetime import datetime, timedelta


# Non-Python dependencies. Not sure if they are installed
# import GTK3
try:
    import gi
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk, GLib, Gdk
except:
    sys.stderr.write("Please install gtk3 and python-gobject, or equivalent",
                     file=sys.stderr)
    sys.exit(1)


# command line parsing to determine the storage provider
parser = argparse.ArgumentParser( prog = "krapplet",
                                  description = "A password manager")

# currently only gnome-keyring, and pass are supported,
# later on sqlite, and keepass might be added
parser.add_argument( "--storage", nargs='?',
                     default="gnome-keyring",
                     help="gnome-keyring, pass, or memory")
parser.add_argument( "--gpg-id", nargs='?',
                     help="required for pass storage provider")
parser.add_argument( "--armor", action='store_true',
                     help="ASCII encode secrets")
cmd_line = parser.parse_args()
from .kr_storage_select import StorageSelect
st_sel=StorageSelect(cmd_line.storage, cmd_line.gpg_id, cmd_line.armor)
from .kr_storage import \
    connection_open, create_keyring, get_all_keyrings, search_keys, \
    NotAvailableException, LockedException, KeyNotFoundException, \
    PromptDismissedException

from .kr_popup import show_error_message, show_info_message
from . import kr_backup

# application constants:
BSD_LICENSE = """
Copyright 2020-2021 Johannes Willem Fernhout <hfern@fernhout.info>.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE."""
APP_RESPONSE_DELETE = 1
FRAME_LABEL_XALIGN = 0.025
APP_EMBED_TIMEOUT = 5000                       # 5 seconds
PW_SPECIAL_CHARS = "!#$%&()*+,-./:;<=>?@[\\]^_`{|}~"
ICONNAME = "krapplet"
DEFAULT_PW_LEN = 16
__VERSION__ = "0.4.0"


# utility functions:
def add_item2menu(mnu=None, label=None, action=None, data=None):
    """ adds a menuitem to a menu (mnu), optionally with an activate action
        and data to be passed to the action """
    if  mnu is None or label is None:
        print("add_item2menu: mnu nor label can be None", file=sys.stderr)
        raise AssertionError
    mni =  Gtk.MenuItem(label=label)
    if action:
        if data:
            mni.connect("activate", action, data)
        else:
            mni.connect("activate", action)
    mnu.append( mni )


def add_separator2menu(mnu):
    """ add_separator2menu: adds a separator to a menu """
    mni =  Gtk.SeparatorMenuItem()
    mnu.append( mni )

"""
def show_info_or_error_message(active_widget, primary_msg,
                               secondary_msg, msg_type, title, buttons ):
    dialog = Gtk.MessageDialog(title=title, transient_for=active_widget,
                               flags=0, message_type=msg_type,
                               buttons=buttons, text=primary_msg)
    if secondary_msg:
        dialog.format_secondary_text(secondary_msg)
    dialog.run()
    dialog.destroy()


def show_error_message(active_widget, primary_msg, secondary_msg):
    show_info_or_error_message(active_widget, primary_msg, secondary_msg,
                               Gtk.MessageType.ERROR, "ERROR",
                               Gtk.ButtonsType.CLOSE)


def show_info_message(active_widget, primary_msg, secondary_msg):
    show_info_or_error_message(active_widget, primary_msg, secondary_msg,
                               Gtk.MessageType.INFO, "Informational",
                               Gtk.ButtonsType.CLOSE)
"""

def abs_path(rel_path):
    """ returns the absolute path for a file,
    given the relative path to this ,py file """
    packagedir = os.path.dirname(__file__)
    joinedpath = os.path.join(packagedir, rel_path)
    path = os.path.abspath(joinedpath)
    return path


def icon_path(iconname, res):
    """ finds the path to an icon, based on std Gtk functions,
    so looking in standard locations like $HOME/.icons,
    $XDG_DATA_DIRS/icons, and /usr/share/pixmaps """
    icon_theme = Gtk.IconTheme.get_default()
    icon = icon_theme.lookup_icon(iconname, res, 0)
    if icon:
        path = icon.get_filename()
        return path
    print("Krapplet icon "
          + iconname
          + " with resolution "
          + str(res)
          + " not found", file=sys.stderr)
    return ""


def copy_text2clipboard(txt):
    """ copy_text2clipboard copies the txt parameter to the cliboard """
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    clipboard.set_text(txt, -1)


def timestamp(since_epoch):
    """returns a string according to local locale
    based on seconds since epoch 1-JAN-1970"""
    if since_epoch:
        from_timestamp = datetime.fromtimestamp(since_epoch)
        date_string = from_timestamp.strftime(
            locale.nl_langinfo(locale.D_T_FMT))
        return date_string
    return ""


def int2(a_string):
    """ converts string to in, return 0 on failure """
    try:
        return int(a_string)
    except ValueError:
        return 0


def sigint_handler(sig, frame):      # frame is stack frame, and is ignored
    """ Signal handler for SIGINT, or Ctrl-C,
    to avoid standard Python stack dump """
    print("Signal", sig, "received, terminating", file=sys.stderr)
    toplevels = Gtk.Window.list_toplevels()
    for toplevel in toplevels:
        toplevel.destroy()
    Gtk.main_quit()


def setup_ccs():
    """ Sets up neccessary cascading style sheets"""
    # FIXME: should come from file, not from memory
    css = """
        levelbar block.weak   { background-color: #FF0000; border: #FF0000; }
        levelbar block.ok     { background-color: #FFA500; border: #FF0000; }
        levelbar block.strong { background-color: #00FF00; border: #FF0000; }
        button:active         { background-color: #0274d9; }
    """
    css_provider = Gtk.CssProvider()
    css_provider.load_from_data(bytes(css.encode()))
    Gtk.StyleContext.add_provider_for_screen(
        Gdk.Screen.get_default(),
        css_provider,
        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    )


def password_complexity( passwd):
    """ Calculates password complexity based on the logic found at this
    website: http://passwordstrengthcalculator.com/interpret.php.
    Complexity is the base 2 log of the cardinalty times the password length"""
    if len(passwd) == 0:
        return 0
    upperc_present = lowerc_present = num_present = special_present = False
    cardinality = 0
    for char in passwd:
        if not upperc_present and char in string.ascii_uppercase:
            cardinality += len(string.ascii_uppercase)
            upperc_present = True
        elif not lowerc_present and char in string.ascii_lowercase:
            cardinality += len(string.ascii_lowercase)
            lowerc_present = True
        elif not num_present and char in string.digits:
            cardinality += len(string.digits)
            num_present = True
        elif not special_present and char in PW_SPECIAL_CHARS:
            cardinality += len(PW_SPECIAL_CHARS)
            special_present = True
    complexity = math.log2(cardinality) * len(passwd)
    return int(complexity)


def gen_pw( tot_len, upperc_len, lowerc_len, numeric_len, special_len):
    """ Returns a generated password based on total lenght, and
    the number of capital, lowercase, numeric and special characters"""

    def part_passwd(length, chars):
        """passwd create string of random chars from a set of chars"""
        passw = ""
        #for i in range(length):
        for _ in range(length):
            passw += secrets.choice(chars)
        return passw

    def shuffle_word(word):
        """shuffle_word shuffles the letters of a word"""
        word = list(word)
        random.shuffle(word)
        return ''.join(word)

    extra_len = tot_len - upperc_len - lowerc_len - numeric_len - special_len
    upperc_str = part_passwd(upperc_len, string.ascii_uppercase)
    lowerc_str = part_passwd(lowerc_len, string.ascii_lowercase)
    numeric_str = part_passwd(numeric_len, string.digits)
    special_str = part_passwd(special_len, PW_SPECIAL_CHARS)
    extra_str = ""
    if extra_len > 0:
        if upperc_len:
            extra_str += string.ascii_uppercase
        if lowerc_len:
            extra_str += string.ascii_lowercase
        if numeric_len:
            extra_str += string.digits
        if special_len:
            extra_str += PW_SPECIAL_CHARS
        extra_str = part_passwd(extra_len, extra_str)
    passw = shuffle_word(upperc_str
                      + lowerc_str
                      + numeric_str
                      + special_str
                      + extra_str)
    return passw


class PasswordGeneratorDialog(Gtk.Dialog):
    """ NewPasswordDialog: dialog window to generate a new password """
    def __init__(self, parent ):
        Gtk.Dialog.__init__(self, title="Generate password", flags=0)
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.tot_len = DEFAULT_PW_LEN
        self.upperc_len = 2
        self.lowerc_len = 2
        self.numeric_len = 2
        self.special_len = 2
        self.show_window()
        self.set_transient_for(parent)

    def show_window( self ):
        """ show_window builds up a window based on the values i
            of self.new_keyname/new_attribs/new_secret """

        self.set_default_size(100, 100)
        box = self.get_content_area()
        frame = Gtk.Frame(label="Parameters", label_xalign=FRAME_LABEL_XALIGN)
        box.add(frame)
        self.grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        frame.add(self.grid)

        prompt_length = Gtk.Label(label="Total length", xalign=0)
        self.entry_lenght = Gtk.Entry( text=str(self.tot_len), \
                                       input_purpose=Gtk.InputPurpose.DIGITS,
                                       xalign=0, width_request=3 )
        self.entry_lenght.connect("insert_text", self.validate, 2)
        prompt_upperc = Gtk.Label(label="Uppercase", xalign=0)
        self.entry_upperc = Gtk.Entry(text=str(self.upperc_len),
                                  input_purpose=Gtk.InputPurpose.DIGITS,
                                  xalign=0)
        self.entry_upperc.connect("insert_text", self.validate, 2)
        prompt_lowerc = Gtk.Label(label="Lowercase", xalign=0)
        self.entry_lowerc = Gtk.Entry(text=str(self.lowerc_len),
                                  input_purpose=Gtk.InputPurpose.DIGITS,
                                  xalign=0 )
        self.entry_lowerc.connect("insert_text", self.validate, 2)
        prompt_numeric = Gtk.Label(label="Numeric", xalign=0)
        self.entry_numeric = Gtk.Entry(text=str(self.numeric_len),
                                  input_purpose=Gtk.InputPurpose.DIGITS,
                                  xalign=0 )
        self.entry_numeric.connect("insert_text", self.validate, 2)
        prompt_special = Gtk.Label(label="special", xalign=0)
        self.entry_special = Gtk.Entry(text=str(self.special_len),
                                  input_purpose=Gtk.InputPurpose.DIGITS,
                                  xalign=0 )
        self.entry_special.connect("insert_text", self.validate, 2)
        empty_line = Gtk.Label()

        self.grid.attach_next_to(prompt_length, None,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.entry_lenght, prompt_length,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.grid.attach_next_to(prompt_upperc, prompt_length,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.entry_upperc, prompt_upperc,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.grid.attach_next_to(prompt_lowerc, prompt_upperc,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.entry_lowerc, prompt_lowerc,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.grid.attach_next_to(prompt_numeric, prompt_lowerc,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.entry_numeric, prompt_numeric,
                                 Gtk.PositionType.RIGHT, 1, 1)
        self.grid.attach_next_to(prompt_special, prompt_numeric,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        self.grid.attach_next_to(self.entry_special, prompt_special,
                                 Gtk.PositionType.RIGHT, 1, 1)
        box.add( empty_line )
        self.show_all()

    def update_from_window( self ):
        """ updates instance vars from window """
        self.tot_len = int2(self.entry_lenght.get_text())
        self.upperc_len = int2(self.entry_upperc.get_text())
        self.lowerc_len = int2(self.entry_lowerc.get_text())
        self.numeric_len = int2(self.entry_numeric.get_text())
        self.special_len = int2(self.entry_special.get_text())

    def validate(self, widget, new_text, new_text_length, position, maxlen ):
        """ validates keyboard input for numberic, and nrof digits, and
            enables/disables the OK button based on the input given"""
        field_contents = widget.get_text()
        newpos = widget.get_position()
        if (new_text.isnumeric()
            and (new_text_length + len(field_contents)) <= maxlen):
            field_contents = (field_contents[:newpos]
                              + new_text
                              + field_contents[newpos:])
        widget.handler_block_by_func(self.validate)
        widget.set_text(field_contents)
        widget.handler_unblock_by_func(self.validate)
        GLib.idle_add(widget.set_position, newpos+1)
        widget.stop_emission_by_name("insert_text")
        self.update_from_window()
        enabled = ((self.tot_len >= (self.upperc_len + self.lowerc_len
                                     + self.numeric_len + self.special_len))
                   and (self.upperc_len or self.lowerc_len
                        or self.numeric_len or self.special_len ))
        self.set_response_sensitive( Gtk.ResponseType.OK, enabled )

    def generate_new_password(self):
        """ Returns a newly generated passwd """
        self.update_from_window()
        passwd = gen_pw(self.tot_len, self.upperc_len,
                        self.lowerc_len, self.numeric_len, self.special_len)
        return passwd


class KeyDialog(Gtk.Dialog):
    """ KeyDialog: shows a dialog window for keys """
    def __init__(self, parent, key=None, keyring=None):
        Gtk.Dialog.__init__(self, title="Edit key", flags=0 )
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_DELETE, APP_RESPONSE_DELETE,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.key = key
        self.keyring = keyring
        self.extra_attribs = 0
        self.out_of_focus_handler = None
        if self.key:
            self.old_keyring_label = self.new_keyring_label = key.keyring_label
            self.old_keyname = self.new_keyname = key.get_label()
            self.old_attribs = self.new_attribs = key.get_attributes()
            self.old_secret = self.new_secret = key.get_secret().decode("utf-8")
            self.old_created = key.get_created()
            self.old_modified= key.get_modified()
            self.connection = self.key.connection
        else:
            self.old_keyring_label = self.new_keyring_label = keyring.get_label()
            self.old_keyname = self.new_keyname = "New key"
            self.old_attribs = self.new_attribs = {}
            self.new_attribs["URL"] = "https://"
            self.new_attribs["username"] = ""
            self.new_attribs["validity"] = "90"
            self.old_secret = self.new_secret = gen_pw(DEFAULT_PW_LEN,
                                                       2, 2, 2, 2)
            self.old_created = self.old_modified = 0
            self.connection = keyring.connection
        self.secret_visibilty = False
        self.box = None
        self.show_window()

    def show_window( self ):
        """ KeyDialog.show_window builds up a window based on the values i
            of self.new_keyname/new_attribs/new_secret """
        self.set_default_size(150, 100)
        content_area = self.get_content_area()
        if self.box:
            self.box.destroy()
        self.box = Gtk.Box(orientation = Gtk.Orientation.VERTICAL)
        content_area.add(self.box)

        keyring_frame = Gtk.Frame(label="Keyring",
                                  label_xalign=FRAME_LABEL_XALIGN)
        keyring_box = Gtk.Box(orientation = Gtk.Orientation.VERTICAL)
        self.box.add(keyring_frame)
        keyring_frame.add(keyring_box)
        self.newkeyring_combo = Gtk.ComboBoxText()
        idx = active_idx = 0
        keyrings = get_all_keyrings(self.connection)
        for keyring in keyrings:
            keyring_label = keyring.get_label()
            if len(keyring_label) > 0:
                self.newkeyring_combo.append_text(keyring_label)
                if keyring_label == self.new_keyring_label:
                    self.keyring = keyring
                    active_idx = idx
                idx += 1
        self.newkeyring_combo.set_active(active_idx)
        keyring_box.pack_start(self.newkeyring_combo, True, True, 0)

        key_frame = Gtk.Frame(label="Key", label_xalign=FRAME_LABEL_XALIGN)
        key_grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        self.box.add(key_frame)
        key_frame.add(key_grid)

        self.newkey_entry = Gtk.Entry(text=self.new_keyname,xalign=0)
        self.newkey_entry.set_max_width_chars(64)
        key_grid.attach_next_to(self.newkey_entry, None,
                                Gtk.PositionType.RIGHT, 2, 1)
        last_prompt = self.newkey_entry

        if self.old_created:
            if self.old_created != self.old_modified:
                created_prompt = Gtk.Label(label="Created", xalign=0)
                created_value  = Gtk.Label(label=timestamp( self.old_created),
                                                            xalign=0)
                key_grid.attach_next_to(created_prompt, last_prompt,
                                        Gtk.PositionType.BOTTOM, 1, 1)
                key_grid.attach_next_to(created_value, created_prompt,
                                        Gtk.PositionType.RIGHT, 1, 1)
                last_prompt = created_prompt
            modified_prompt = Gtk.Label(label=" Modified", xalign=0)
            modified_value  = Gtk.Label(label=timestamp( self.old_modified ),
                                                         xalign=0)
            key_grid.attach_next_to(modified_prompt, last_prompt,
                                    Gtk.PositionType.BOTTOM, 1, 1)
            key_grid.attach_next_to(modified_value, modified_prompt,
                                    Gtk.PositionType.RIGHT, 1, 1)
            last_prompt = modified_prompt

        self.box.add( Gtk.Label())
        attr_frame = Gtk.Frame(label="Attributes",
                               label_xalign=FRAME_LABEL_XALIGN)
        attr_grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        self.box.add(attr_frame)
        attr_frame.add(attr_grid)
        self.attr_prompt = {}
        self.attr_value  = {}
        last_prompt = None
        launch_button_active = False
        expiry_check = False
        for attr in self.new_attribs:
            if  attr != "xdg:schema":
                self.attr_prompt[attr] = Gtk.Entry(text=attr, xalign=0,
                                                   max_width_chars=20)
                self.attr_prompt[attr].set_max_width_chars(20)
                self.attr_value[attr]  = Gtk.Entry(text=self.new_attribs[attr],
                                                   xalign = 0)
                self.attr_value[attr].set_max_width_chars( 42 )
                attr_grid.attach_next_to(self.attr_prompt[attr], last_prompt,
                                         Gtk.PositionType.BOTTOM, 1, 1 )
                attr_grid.attach_next_to(self.attr_value[attr],
                                         self.attr_prompt[attr],
                                         Gtk.PositionType.RIGHT, 1, 1)
                if attr == "URL":
                    launch_button_active = True
                if attr == "validity":
                    expiry_check = self.old_created
                last_prompt = self.attr_prompt[attr]
                last_entry  = self.attr_value[attr]
        if self.extra_attribs:
            self.extra_attr_prompt = Gtk.Entry(xalign = 0)
            self.extra_attr_prompt.set_max_width_chars(20)
            self.extra_attr_entry  = Gtk.Entry(xalign = 0)
            self.extra_attr_entry.set_max_width_chars(42)
            attr_grid.attach_next_to(self.extra_attr_prompt, last_prompt,
                                     Gtk.PositionType.BOTTOM, 1, 1 )
            attr_grid.attach_next_to(self.extra_attr_entry,
                                     self.extra_attr_prompt,
                                     Gtk.PositionType.RIGHT, 1, 1 )
            last_prompt = self.extra_attr_prompt
            last_entry  = self.extra_attr_entry
        attr_button_box = Gtk.Box()
        if launch_button_active:
            launch_button = Gtk.Button(label="Launch")
            launch_button.connect("clicked", self.launch)
            attr_button_box.pack_end(launch_button, False, False, 0)
        add_button = Gtk.Button(label="Add")
        add_button.connect("clicked", self.add_attr)
        attr_button_box.pack_end(add_button, False, False, 2 )
        attr_grid.attach_next_to(attr_button_box, last_prompt,
                                 Gtk.PositionType.BOTTOM, 2, 1)
        self.box.add( Gtk.Label())
        label=secr_frame_label = "Secret"
        expiring = expired = False
        if expiry_check:
            validity = int(self.new_attribs["validity"])
            now = datetime.now()
            modified = datetime.fromtimestamp(self.old_modified)
            expired_date  = modified + timedelta(days=validity)
            expiring_date = modified + timedelta(days=validity - 7)
            if expired_date  < now:
                expired = True
                label = '<span foreground="red" weight="bold">Expired secret</span>'
                secr_frame_label = label
            elif expiring_date < now:
                expiring = True
                label = '<span foreground="orange" weight="bold">Expiring secret</span>'
                secr_frame_label = label
        secr_frame = Gtk.Frame(label=secr_frame_label,
                               label_xalign=FRAME_LABEL_XALIGN)
        if expiring or expired:
            secr_frame.get_label_widget().set_use_markup(True)
        secr_grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        self.box.add(secr_frame)
        secr_frame.add(secr_grid)
        self.secret_value = Gtk.Entry(text=self.new_secret, xalign=0,
                                      visibility=self.secret_visibilty)
        self.secret_value.connect("changed", self.update_complexity_levelbar)
        self.secret_value.set_max_width_chars(64)
        self.complexity_levelbar = Gtk.LevelBar()
        self.complexity_levelbar.set_min_value(password_complexity("abcdefgh"))
        self.complexity_levelbar.set_max_value(100)
        self.complexity_levelbar.add_offset_value("weak", password_complexity("Aa!456789"))
        self.complexity_levelbar.add_offset_value("ok", password_complexity("Aa!45678901"))
        self.complexity_levelbar.add_offset_value("strong", 100)
        self.complexity_levelbar.set_value( min(100, password_complexity( self.new_secret )))
        secret_copy_button = Gtk.Button(label="Copy")
        secret_copy_button.connect("clicked", self.copy_secret2clipboard)
        visi_text = "Hide" if self.secret_visibilty else "Show"
        visi_button = Gtk.Button(label=visi_text)
        visi_button.connect("clicked", self.toggle_secret_visibility)
        gen_button = Gtk.Button(label="Generate")
        gen_button.connect("clicked", self.generate_password)
        secret_promptbox = Gtk.Box(spacing=2)
        secret_promptbox.pack_end(visi_button, False, False, 0)
        secret_promptbox.pack_end(secret_copy_button, False, False, 0)
        secret_promptbox.pack_end(gen_button, False, False, 0)
        secr_grid.attach_next_to(self.secret_value, None,
                                 Gtk.PositionType.RIGHT, 1, 1)
        secr_grid.attach_next_to(self.complexity_levelbar, self.secret_value,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        secr_grid.attach_next_to(secret_promptbox, self.complexity_levelbar,
                                 Gtk.PositionType.BOTTOM, 1, 1)
        last_prompt = secret_promptbox
        emptyline = Gtk.Label()
        self.box.add( emptyline )
        self.show_all()

    def update_complexity_levelbar(self, widget):
        """updates the complexity indicator when entry
        is done in the secret field"""
        pw = widget.get_text()
        complexity = password_complexity(pw)
        self.complexity_levelbar.set_value(min(100, complexity))

    def generate_password(self, button):
        """ action function for generate password """
        self.update_key_from_window()
        self.active_widget = dialog = PasswordGeneratorDialog(self)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.new_secret = dialog.generate_new_password()
        dialog.destroy()
        self.active_widget = self
        self.secret_visibilty = False
        self.show_window()

    def launch(self, button):
        """ tries to find an attrib["URL"] and lauches it in a webbrowser """
        self.update_key_from_window()
        url = self.new_attribs["URL"]
        webbrowser.open(url, new=0, autoraise=True)
        self.out_of_focus_handler = self.connect("focus-out-event",
                                                 self.refocus)

    def refocus(self, widget, event):
        """ Tries to grab focus back after we lost it due to lauching an URL """
        self.present()
        if self.out_of_focus_handler:
            self.disconnect(self.out_of_focus_handler)
            self.out_of_focus_handler = None

    def update_key_from_window(self):
        """ copies the values from screen to internal structures """
        self.new_keyring_label = self.newkeyring_combo.get_active_text()
        self.new_keyname = self.newkey_entry.get_text()
        screen_attribs = {}
        for attr in self.new_attribs:
            if  attr != "xdg:schema":
                prompt = self.attr_prompt[attr].get_text()
                if prompt != "":
                    value  = self.attr_value[attr].get_text()
                    screen_attribs[prompt] = value
        if self.extra_attribs:
            prompt = self.extra_attr_prompt.get_text()
            if prompt != "":
                value  = self.extra_attr_entry.get_text()
                screen_attribs[prompt] = value
            self.extra_attribs = False
        self.new_secret = self.secret_value.get_text()
        self.new_attribs = screen_attribs

    def toggle_secret_visibility(self, button):
        """ toggles the visibiity of the password """
        self.update_key_from_window()
        self.secret_visibilty = not self.secret_visibilty
        self.show_window()

    def add_attr(self, button):
        """ adds an attribute line to the window """
        self.update_key_from_window()
        self.extra_attribs = True
        self.show_window()

    def copy_secret2clipboard(self, widget):
        " copies the secret from screen to the clipboard """
        self.update_key_from_window()
        copy_text2clipboard(self.new_secret)

    def deletekey(self):
        """ Deletes a key """
        dialog = Gtk.Dialog(title="Delete key", flags=0)
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        dialog.set_default_size(150, 100)
        content_area = dialog.get_content_area()
        box = Gtk.Box(orientation = Gtk.Orientation.VERTICAL)
        box.pack_start(Gtk.Label(), True, True, 0)
        box.pack_start(Gtk.Label(label="  Are you sure to delete key '"
                                + self.old_keyname
                                + "' ?  "), True, True, 0)
        content_area.add(box)
        dialog.show_all()
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.key.delete()
        dialog.destroy()

    def savekey(self):
        """ Saves a key """
        self.update_key_from_window()
        if self.old_keyname != self.new_keyname:
            self.key.set_label(self.new_keyname)
        if self.old_attribs != self.new_attribs:
            self.key.set_attributes(self.new_attribs)
        if self.old_secret != self.new_secret:
            self.key.set_secret(self.new_secret.encode("utf-8"))
        if self.keyring and self.old_keyring_label != self.new_keyring_label:
            self.key.move(self.new_keyring_label)

    def create_newkey( self, keyring):
        """ Creates a new key """
        self.update_key_from_window()
        self.key = keyring.create_key( label=self.new_keyname,
                                       attributes=self.new_attribs,
                                       secret=self.new_secret.encode("utf-8"))


class AddKeyRingDialog(Gtk.Dialog):
    """ AddKeyRingDialog: window class to add a keyring,
        prompts for keyring name """
    def __init__(self, parent):
        Gtk.Dialog.__init__(self, title="New keyring", flags=0)
        self.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                         Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.set_default_size(150, 100)
        box = self.get_content_area()
        keyringname_prompt = Gtk.Label(label="Keyring name", xalign=0)
        self.keyringname_input = Gtk.Entry(text="New keyring", xalign = 0)
        grid = Gtk.Grid(row_spacing=2, column_spacing=5)
        box.add(grid)
        grid.attach_next_to(keyringname_prompt, None,
                            Gtk.PositionType.BOTTOM, 1, 1)
        grid.attach_next_to(self.keyringname_input, keyringname_prompt,
                            Gtk.PositionType.RIGHT, 1, 1)
        self.set_position(Gtk.WindowPosition.MOUSE)
        self.show_all()

    def get_newkeyringname(self):
        """ get_newkeyringname: returns the entered name for a new keyring"""
        return self.keyringname_input.get_text()


class StatusIcon:
    """Main entry point, creates a tray icon, and handles rightclicks"""
    def __init__(self):
        self.dbus_open = False
        self.statusicon = Gtk.StatusIcon()
        self.statusicon.set_from_file( icon_path(ICONNAME, 48))
        self.statusicon.connect("popup-menu", self.right_click_event)
        self.statusicon.set_title("Password manager")
        self.statusicon.set_tooltip_text("krapplet: a password manager")
        self.statusicon.set_visible(True)
        self.embedded = False
        GLib.idle_add(self.embed_check)
        GLib.timeout_add(APP_EMBED_TIMEOUT, self.embed_timeout)
        self.active_widget = None

    def embed_check(self):
        """ keep calling until we are embedded """
        if self.statusicon.is_embedded():
            self.embedded = True
            return False                    # don't call again
        return True

    def embed_timeout(self):
        """ if still not emebedded then exit program """
        if self.embedded:
            return False                    # don't call again
        print("Error: could not embed in systray", file=sys.stderr)
        Gtk.main_quit()
        return True

    def right_click_event(self, icon, button, time):
        """ shows the popop menu """
        if self.active_widget:
            self.active_widget.present()
            return
        if self.dbus_open:
            self.connection.close()
        self.connection = connection_open()
        self.dbus_open = True
        self.menu = Gtk.Menu()
        self.keyrings = get_all_keyrings(self.connection)
        self.keyring_labels = []
        for keyring in self.keyrings:
            keyring_label = keyring.get_label()
            if len(keyring_label):
                item = Gtk.MenuItem(label=keyring_label)
                submenu = Gtk.Menu()
                if keyring.is_locked():             # try auto unlock
                    unlock_searchkey = {"keyring" : "LOCAL:/keyrings/"
                                        + keyring_label.replace(" ", "_")
                                        + ".keyring"}
                    unlock_keys = search_keys(self.connection,
                                              unlock_searchkey)
                    for unlock_key in unlock_keys:
                        if not unlock_key.is_locked():
                            keyring.unlock()
                if keyring.is_locked():
                    add_item2menu(mnu=submenu, label="Unlock",
                                  action=self.unlock_keyring, data=keyring)
                else:
                    nrof_keys = 0
                    self.keyring_labels.append(keyring_label)
                    try:
                        for key in keyring.get_all_keys():
                            if key.is_locked():
                                lock_return = key.unlock()
                            key.keyring_label = keyring_label
                            key.keyring_labels = self.keyring_labels
                            add_item2menu(mnu=submenu, label=key.get_label(),
                                          action=self.key_dialog, data=key)
                            nrof_keys += 1
                    except KeyNotFoundException:
                        show_error_message(
                            self.active_widget,
                            "Key not found exception",
                            "Restarting gnome-keyring; this will lock all keyrings" )
                        msg = subprocess.check_output(
                            ["gnome-keyring-daemon", "-r", "-d"],
                            stderr=subprocess.STDOUT )
                        show_info_message(
                             self.active_widget,
                             "Output of gnome-restart command",
                             msg.decode("utf-8"))
                    except:
                        print( "class sys.exc_info()[0]:" , sys.exc_info()[0].__class__)
                        """show_error_message(
                             self.active_widget,
                             "Unexpected error occured, please report this error",
                             sys.exc_info()[0]) """
                    if nrof_keys:
                        add_separator2menu(submenu)
                    add_item2menu(mnu=submenu, label="Add key",
                                  action=self.add_key, data=keyring)
                    if  not nrof_keys:
                        add_item2menu(mnu=submenu, label="Remove keyring",
                                      action=self.remove_keyring, data=keyring)
                    add_item2menu(mnu=submenu, label="Lock keyring",
                                  action=self.lock_keyring, data=keyring )
                item.set_submenu(submenu)
                self.menu.append(item)
        add_separator2menu( self.menu )
        add_item2menu(mnu=self.menu, label="Add keyring",
                      action=self.add_keyring, data=None)
        add_item2menu(mnu=self.menu, label="Backup",
                      action=self.backup, data=None)
        add_item2menu(mnu=self.menu, label="Restore",
                      action=self.restore, data=None)
        add_item2menu(mnu=self.menu, label="About",
                      action=self.show_about_dialog, data=None)
        add_item2menu(mnu=self.menu, label="Quit",
                      action=Gtk.main_quit, data=None)
        self.menu.show_all()
        self.menu.popup(None, None, None, self.statusicon, button, time)

    def unlock_keyring(self, widget, keyring):
        """Unocks a keyring"""
        keyring.unlock()

    def lock_keyring(self, widget, keyring):
        """Locks a keyring"""
        keyring.lock()

    def remove_keyring(self, widget, keyring):
        """Removes a keyring, when it is empty.
        If not it shows an error message"""
        keycount = 0
        for key in keyring.get_all_keys():
            keycount += 1
        if keycount:
            show_error_message(self.active_widget,
                               "Error deleting  keyring",
                               "Keys are still attached to keyring" )
        else:
            keyring.delete()

    def key_dialog(self, widget, key):
        """Pops up a window showing a key, with all the attribs"""
        if not key.is_locked():               # check check double check
            self.active_widget = dialog = KeyDialog(self, key=key)
            response = dialog.run()
            if response == Gtk.ResponseType.OK:
                dialog.savekey()
            elif response == APP_RESPONSE_DELETE:
                dialog.deletekey()
            copy_text2clipboard("")
            dialog.destroy()
            self.active_widget = None
        else:
            show_error_message(self.active_widget,
                               "Key locked", key.get_label() )

    def add_key(self, widget, keyring):
        """Add_key adds a key to a keyrng"""
        self.active_widget = dialog = KeyDialog(self, keyring=keyring)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            dialog.create_newkey(keyring)
        dialog.destroy()
        self.active_widget = None

    def add_keyring(self, widget):
        """Pops up a window asking for a new keyring name"""
        self.active_widget = dialog = AddKeyRingDialog(self)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            keyringname = dialog.get_newkeyringname()
            try:
                create_keyring(self.connection, keyringname)
            except Exception as e:
                show_error_message(self.active_widget,
                                   "Error saving new keyring", str( e ))
        dialog.destroy()
        self.active_widget = None

    def backup(self, widget):
        kr_backup.backup()

    def restore(self, widget):
        kr_backup.restore()

    def show_about_dialog(self, widget):
        """Shows the about dialog"""
        image = Gtk.Image()
        image.set_from_file(icon_path( ICONNAME, 96))
        icon_pixbuf = image.get_pixbuf()
        self.active_widget = about_dialog = Gtk.AboutDialog()
        about_dialog.set_destroy_with_parent(True)
        about_dialog.set_logo(icon_pixbuf)
        about_dialog.set_program_name("krapplet")
        about_dialog.set_version("Version " + __VERSION__)
        about_dialog.set_comments("A password manager using gnome-keyring or pass")
        about_dialog.set_authors(["Johannes Willem Fernhout"])
        about_dialog.set_copyright( "(c) 2020-2022 Johannes Willem Fernhout")
        about_dialog.set_license(BSD_LICENSE)
        about_dialog.set_website("https://gitlab.com/hfernh/krapplet")
        about_dialog.set_website_label("krapplet on GitLab")
        about_dialog.show_all()
        about_dialog.run()
        about_dialog.destroy()
        self.active_widget = None

    def __del__(self):
        """ exit code, just some housekeeping """
        if self.dbus_open:
            self.connection.close()


def main():
    """ krapplet main """
    # ignore GTK deprecation warnings gwhen not in development mode
    # for development mode, run program as python3 -X dev krapplet
    if not sys.warnoptions:
        import warnings
        warnings.simplefilter("ignore")

    setup_ccs()
    ico = StatusIcon()
    signal.signal(signal.SIGINT, sigint_handler)
    Gtk.main()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
kr_crypt: functions for encrypting and decryting strings
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

import sys
import base64
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def passwd_in_bytes(passphrase):
    if passphrase:
        if isinstance(passphrase, bytes):
            return passphrase
        if isinstance(passphrase, str):
            return str.encode(passphrase)
        else:
            print("unexpected type for passphrase", type(passphrase))
            return b''
    return None

def fernet(passphrase):
    """ Returns a Fernet object, based on a passwd (as a string) """
    salt = "[zG?+=-$xD\S{*A#1-.l^&]z@BL;Z:i%".encode()
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32,
                     salt=salt, iterations=100000)
    key = base64.urlsafe_b64encode(kdf.derive(passwd_in_bytes(passphrase)))
    return Fernet(key)

def sym_encrypt(unenc_str, passphrase):
    """ returns a password protected encrypted string """
    fern = fernet(passphrase)
    if isinstance(unenc_str, bytes):
        unenc_bytes = unenc_str
    elif isinstance(unenc_str, str):
        unenc_bytes = str.encode(unenc_str)
    else:
        print("sym_encrpt: cannot convert", type(unenc_str), "to bytes")
    return fern.encrypt(unenc_bytes)

def sym_decrypt( enc_str, passphrase):
    """ returns a decrypted string """
    fern = fernet(passphrase)
    #try:
    if True:
        if isinstance(enc_str, str):
            enc_str = enc_str.encode()
        dec_str = fern.decrypt(enc_str)
        return dec_str
    """
    except InvalidToken:
        #print( "decrypt failed: invalid token", file=sys.stderr)
        pass
    except:
        print("Unexpected error:", sys.exc_info()[0], file=sys.stderr)"""
    return None



from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, padding
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.asymmetric.utils import \
    decode_dss_signature, encode_dss_signature
from cryptography.exceptions import InvalidSignature

def write_serkey(serkey, fname):
    "Writes a serialized key to a file with name fname"
    with fopen(fname, "wb") as fh:
        fh.write(serkey)

def read_serkey(fname):
    "Reads serialized key from a file"
    serkey = b''
    try: 
        with fopen(fname, "rb") as fh:
            fh.read(serkey)
        return serkey
    except Exception as exc:
        print("Can't read seralized key from file", fname, ":", exc)
    return None

def asym_serialize_privkey(privkey, passphrase, fname=None):
    """Serializes a key for storage.
    Will write the serialized key to file fname if fname is not None"""
    if passphrase:
        algo = serialization.BestAvailableEncryption(
            passwd_in_bytes(passphrase))
    else:
        algo = serialization.NoEncryption()
    serialized_privkey = privkey.private_bytes(
        encoding=serialization.Encoding.PEM,
        format = serialization.PrivateFormat.PKCS8,
        encryption_algorithm = algo)
    if fname:
        write_serkey(serialized_privkey)
    return serialized_privkey

def asym_deserialize_privkey(ser_privkey, passphrase, fname=None):
    """Deserializes private key in PEM format with optional password.
    Will read the serizalized key from file if fname is not None"""
    if fname:
        ser_privkey = read_serkey(fname)
    if ser_privkey:
        try:
            privkey = serialization.load_pem_private_key(
                ser_privkey, password=passwd_in_bytes(passphrase))
            return privkey
        except ValueError:              # bad passwd
            pass
        except Exception as exc:
            print("Failed to deserialize private key:", exc)
    return None

def asym_serialize_pubkey(pubkey, fname = None):
    "Serialized a pubkey to PEM format"
    ser_pubkey = pubkey.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo)
    if fname:
        write_serkey(ser_pubkey, fname)
    return ser_pubkey

def asym_deserialize_pubkey(ser_pubkey, fname = None):
    "Deserializes a public key in PEM format"
    if fname:
        ser_pubkey = read_serkey(fname)
    if ser_pubkey:
        pubkey = serialization.load_pem_public_key(ser_pubkey)
    return pubkey

def asym_genkey():
    " Generates a private key for asymetric encryption "
    return ec.generate_private_key(ec.SECP384R1())

def asym_pubkey(privkey):
    " Returns the public counterpart of a private key "
    return privkey.public_key()

def asym_encrypt(key, msg):
    "Encrypts a message with a private or public key"
    pad = padding.OAEP(
        mgf=padding.MGF1(algorithm=hashes.SHA256()),
        algorithm=hashes.SHA256(),
        label=None)
    return key.encrypt(msg, pad)

def asym_decrypt(key, enc_msg):
    "Decrypts a message a public or private key"
    pad = padding.OAEP(
        mgf=padding.MGF1(algorithm=hashes.SHA256()),
        algorithm=hashes.SHA256(),
        label=None)
    try:
        msg = key.decrypt(enc_msg, pad)
    except Exception as exc:
        print("Asym_decrypt failed: ", exc)
        msg = None
    return msg

def asym_sharedkey(privkey, peer_pubkey):
    " Returns the shared key for an exchange "
    return privkey.exchange(ec.ECDH(), peer_pubkey)

def asym_derivedkey(key, salt, handshake_data):
    "Creates a derrived key using the handshake data"
    #derived_key = HKDF(algorithm=hashes.SHA256(), length=32, salt=None,
    derived_key = HKDF(algorithm=hashes.SHA256(), length=32, salt=salt,
                       info=b'handshake data').derive(key)
    return derived_key

def asym_derived_sharedkey(privkey, peer_pubkey, salt, handshake_data):
    "Combines the fuctions of asym_sharedkey and asym_derivedkey"
    sharedkey = asym_sharedkey(privkey, peer_pubkey)
    return asym_derivedkey(sharedkey, salt, handshake_data)

def asym_signature(unenc_str, privkey):
    " Signs a message with a private key "
    signature = privkey.sign(unenc_str, ec.ECDSA(hashes.SHA256()))
    dec_dss_sig = decode_dss_signature(signature)
    return dec_dss_sig

def asym_signature_verify(unenc_str, sig, pubkey):
    " Verifies the signature of a message: True when passes "
    enc_dss_sig = encode_dss_signature(sig[0],sig[1])
    try:
        pubkey.verify(enc_dss_sig, unenc_str, ec.ECDSA(hashes.SHA256()))
        return True
    except InvalidSignature:
        return False
    except Exception as exc:
        print("Unexpected error verifying signature:", str(exc))
    return False



#!/usr/bin/env python

"""
kr_popup: functions for showing info or erorr message
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies.
"""

import sys

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib, Gdk


def show_info_or_error_message(active_widget, primary_msg,
                               secondary_msg, msg_type, title, buttons ):
    """Shows an error or info  message dialog"""
    dialog = Gtk.MessageDialog(title=title, transient_for=active_widget,
                               flags=0, message_type=msg_type,
                               buttons=buttons, text=primary_msg)
    if secondary_msg:
        if isinstance(secondary_msg, list):
            secmsg = '\n'.join([ln for ln in secondary_msg])
        elif isinstance(secondary_msg, str):
            secmsg = secondary_msg
        else:
            print("show_info_or_error_message:"
                  + " secondary message is list nor str",
                  file=sys.stderr)
        #dialog.format_secondary_text(secondary_msg)
        dialog.format_secondary_text(secmsg)
    dialog.run()
    dialog.destroy()


def show_error_message(active_widget, primary_msg, secondary_msg):
    """Shows an error message dialog"""
    show_info_or_error_message(active_widget, primary_msg, secondary_msg,
                               Gtk.MessageType.ERROR, "ERROR",
                               Gtk.ButtonsType.CLOSE)


def show_info_message(active_widget, primary_msg, secondary_msg):
    """ Shows an informational message dialog """
    show_info_or_error_message(active_widget, primary_msg, secondary_msg,
                               Gtk.MessageType.INFO, "Informational",
                               Gtk.ButtonsType.CLOSE)




"""
kr_storage_select: part of krapplet, a password manager
This module selects the right stoage provider
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies
"""

import sys

STORAGE_PROVIDERS = ["gnome-keyring", "memory", "pass"]


class StorageSelect():
    """ class to hoild the selected provider in a class variable
        the intended use is that after the commenadline parameter
        evaluation a storage provider is chosen, e.g. "pass", or
        "gnome-keyring", and that is stored here using the constructor.
        Next using the get function the set provider can be retrieved. """

    provider = None
    gpg_id = None
    armor = None

    def __init__(self, provider = None, gpg_id = None, armor = None):
        if provider:
            if provider in STORAGE_PROVIDERS:
                StorageSelect.provider = provider
                if gpg_id:
                    StorageSelect.gpg_id = gpg_id
                if armor:
                    StorageSelect.armor = armor
            else:
                print("krapplet: error: unrecognized storage provider",
                      provider, file=sys.stderr)
                sys.exit(1)



#!/usr/bin/env python3
"""
kr_storage: part of krapplet, a password manager
This is an intermediary layer to distinghuish the different storage providers
Make sure to import kr_storage_select prior to importing this module.
(c) 2020-2023 Johannes Willem Fernhout, BSD 3-Clause License applies
"""

from .kr_storage_select import StorageSelect

st_sel=StorageSelect()
STORAGE_PROVIDER = st_sel.provider

# import storage backends
if STORAGE_PROVIDER == "gnome-keyring":
    # print("using gnome-keyring storage provider")
    from .kr_secretstorage import \
        connection_open, create_keyring, get_all_keyrings, search_keys, \
        NotAvailableException, LockedException, KeyNotFoundException, \
        PromptDismissedException
elif STORAGE_PROVIDER == "pass":
    # print("using pass storage provider")
    from .kr_pass import \
        connection_open, create_keyring, get_all_keyrings, search_keys, \
        NotAvailableException, LockedException, KeyNotFoundException, \
        PromptDismissedException, check_gpg, set_armor
    check_gpg(st_sel.gpg_id)
    set_armor(st_sel.armor)
elif STORAGE_PROVIDER == "memory":
    from .kr_memstorage import \
        connection_open, create_keyring, get_all_keyrings, search_keys, \
        NotAvailableException, LockedException, KeyNotFoundException, \
        PromptDismissedException
